import AlteryxPythonSDK as Sdk
from typing import Union, Any

import azure.datalake.store as adl_store
from adal import AdalError

import connectors_core.alteryx as alteryx
import connectors_core.azure as core_azure
import connectors_core.azure.filewriters as filewriters
from connectors_core.file import FileFormat, FileExistsBehavior
from connectors_core.file.interfaces import FileWriter
from connectors_core.file.errors import FileOpenError, FileWriterAbortedError


class AyxPlugin(alteryx.AyxPluginBase):

    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)

        # Configuration Options
        self.adl_auth_type = core_azure.AzureAuthenticationType.ServicePrincipal
        self.adl_store_name = None  # type: str
        self.adl_tenant_id = None  # type: str
        self.adl_client_id = None  # type: str
        self.adl_secret_key = None  # type: str
        self.adl_username = None  # type: str
        self.adl_password = None  # type: str
        self.adl_file_path = ''
        self.adl_file_format = FileFormat.Delimited
        self.adl_delimited_has_header = True
        self.adl_existing_file_behavior = FileExistsBehavior.Abort

        self.adl_authentication = None
        self.incoming_connection_added = False

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[alteryx.IncomingConnectionBase, None]:
        if self.incoming_connection_added:
            self.display_error_msg("Only one incoming connection can be accepted by this tool")
            return None
        connection_handler = AzureDataLakeIncomingConnectionHandler(self)
        self.incoming_connection_added = True
        return connection_handler

    def pi_init(self, str_xml: str):
        super().pi_init(str_xml)

        # Getting the user-entered file name string from the GUI, and the output anchor from the XML file.
        self.adl_store_name = self.get_xml_config_value(str_xml, 'adl_StoreName')
        self.adl_auth_type = core_azure.AzureAuthenticationType[
            self.get_xml_config_value(str_xml, 'adl_AuthenticationType')
            or core_azure.AzureAuthenticationType.Unknown.name
            ]
        self.adl_tenant_id = self.get_xml_config_value(str_xml, 'adl_TenantId') or ''

        # Validation
        if not self.adl_store_name:
            self.invalidate_config(
                "Please provide your Azure Data Lake Store Name.")
        if self.adl_auth_type == core_azure.AzureAuthenticationType.Unknown:
            self.invalidate_config("Please specify an Authentication Type")
        if not self.adl_tenant_id:
            self.invalidate_config("Please provide a Tenant Id")

        if self.adl_auth_type == core_azure.AzureAuthenticationType.ServicePrincipal:
            self.adl_client_id = self.get_xml_config_value(str_xml, 'adl_ClientId') or ''
            self.adl_secret_key = self.get_encrypted_xml_config_value(str_xml, 'adl_SecretKey') or ''

            # Validation
            if not self.adl_client_id:
                self.invalidate_config("Please provide a Client Id")
            if not self.adl_secret_key:
                self.invalidate_config("Please provide a Secret Key")

            if self.is_initialized:
                self.adl_authentication = core_azure.AdlSpAuthentication(
                    store_name=self.adl_store_name,
                    tenant_id=self.adl_tenant_id,
                    client_id=self.adl_client_id,
                    secret_key=self.adl_secret_key
                )
        elif self.adl_auth_type == core_azure.AzureAuthenticationType.EndUser:
            self.adl_username = self.get_xml_config_value(str_xml, 'adl_UserName')
            self.adl_password = self.get_encrypted_xml_config_value(str_xml, 'adl_Password')

            if not self.adl_username:
                self.invalidate_config("Please provide a User Name")
            if not self.adl_password:
                self.invalidate_config("Please provide a Password")

            if self.is_initialized:
                self.adl_authentication = core_azure.AdlEndUserAuthentication(
                    tenant_id=self.adl_tenant_id,
                    username=self.adl_username,
                    password=self.adl_password)

        else:
            self.invalidate_config("Unknown Authentication Type!")
            return

        # File Configuration
        self.adl_file_format = FileFormat[
            self.get_xml_config_value(str_xml, 'adl_FileFormat') or FileFormat.Unknown.name
            ]
        self.adl_file_path = self.get_xml_config_value(str_xml, 'adl_FilePath')

        self.adl_existing_file_behavior = FileExistsBehavior[
            self.get_xml_config_value(str_xml, 'adl_ExistingFileBehavior') or FileExistsBehavior.Abort.name
        ]

        if not self.adl_file_path:
            # TODO: enforce starting file paths with / ?
            self.invalidate_config("Please provide a valid File Path")
        if self.adl_file_format == FileFormat.Unknown:
            self.invalidate_config("Please specify a File Format")

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        self.display_error_msg("This tool does not provide an output connection")
        return False

    def pi_push_all_records(self, n_record_limit: int) -> bool:
        # We do not have an output to push records to, we are an input-only tool
        return True


class AzureDataLakeIncomingConnectionHandler(alteryx.IncomingConnectionBase):

    def __init__(self, parent: AyxPlugin):
        super().__init__(parent)
        self.writer = None  # type: FileWriter

    def ii_init(self, record_info_in: Sdk.RecordInfo) -> bool:
        if not self.parent.is_initialized:
            return False
        try:
            credentials = self.parent.adl_authentication.get_credentials()
            adl_fs = adl_store.AzureDLFileSystem(store_name=self.parent.adl_store_name, token=credentials)
            file_format = self.parent.adl_file_format

            def conversion_error_handler(field: Sdk.Field, record: Sdk.RecordRef, e: BaseException) -> Any:
                self.parent.display_error_msg(
                    "Data conversion error: encountered for field {} ({}): {}".format(field.name, field.type, e)
                )
                return None

            if file_format == FileFormat.Delimited:
                self.writer = filewriters.ADLDelimitedFileWriter(
                    adl_fs=adl_fs,
                    file_path=self.parent.adl_file_path,
                    write_header=self.parent.adl_delimited_has_header,
                    data_conversion_error_handler=conversion_error_handler,
                    file_exists_behavior=self.parent.adl_existing_file_behavior
                )
            elif file_format == FileFormat.Avro:
                self.writer = filewriters.ADLAvroFileWriter(
                    adl_fs=adl_fs,
                    file_path=self.parent.adl_file_path,
                    data_conversion_error_handler=conversion_error_handler,
                    file_exists_behavior=self.parent.adl_existing_file_behavior
                )
            elif file_format == FileFormat.JSON:
                self.writer = filewriters.ADLJSONFileWriter(
                    adl_fs=adl_fs,
                    file_path=self.parent.adl_file_path,
                    data_conversion_error_handler=conversion_error_handler,
                    file_exists_behavior=self.parent.adl_existing_file_behavior
                )
            else:
                # Unknown file format!
                self.is_initialized = False
                return False
            self.writer.init(record_info_in)
            self.is_initialized = True
            return True
        except ConnectionError as e:
            self.is_initialized = False
            self.parent.display_error_msg("Failed to initialize ADL Output: {}".format(e))
        except FileOpenError as e:
            self.is_initialized = False
            self.parent.display_error_msg("Failed to initialize ADL Output: {}".format(e))
        except AdalError as e:
            self.is_initialized = False
            self.parent.display_error_msg("Failed to authenticate with Azure Data Lake: {}".format(e))

    def ii_push_record(self, in_record: Sdk.RecordRef) -> bool:
        try:
            if not self.writer.is_open():
                self.writer.open()
            return self.writer.write(in_record)
        except FileOpenError as e:
            self.parent.display_error_msg("Failed to open file for writing: {}".format(e))
            return False
        except FileWriterAbortedError as e:
            self.parent.display_error_msg("The write operation was aborted: {}".format(e))
            return False

    def ii_update_progress(self, d_percent: float):
        self.parent.alteryx_engine.output_tool_progress(self.parent.n_tool_id, d_percent)

    def ii_close(self):
        self.writer.close()
        self.writer = None
        self.is_initialized = False
